/* interp.c    interpreter

	som-v16 add file i/o 			7 Nov 2005
	t-code som-v17					1 Jan 2006 (Happy new year!)
	add syscall stop				20 Jan 2006
	som-v19v    tx-code				20 Apr 2007
	som-v5		t2-code				7 Oct 2009

*/

#include "compile.h"

/* run time environment
   define code, data, stack segments here

   system area  1 .. MAXSYS-1
   1..63           local
   100..300        link vectors and string
   390..700        small constants -10..300
   1000... MAXDS   global
*/

int M[MAXMEM];					// data and code segment
int SS[MAXSS];					// stack segment
int *CS = &M[MAXDS];			// code segment
int DS = MAXSYS;				// current free data segment

extern int call_load;			// ads for "call loadfile"
extern int cs_ads;				// base ads for CS[]

PRIVATE FILE *filep[MAXFILE]; 	// store file pointer
PRIVATE int fp = 1;				// current state of computation
PRIVATE int sp = 1;
PRIVATE int param[100];
PRIVATE char strbuf[255];		// buffer for fprints

// switch CS between user and system
#define  user_cs()  CS = &M[M[cs_ads]]
#define  sys_cs()   CS = &M[MAXDS]

int noi = 0;

/*
#ifdef PROF

int icnt[100], lcnt[10000];

void stat(int ip){
	icnt[CS[ip] & 255]++;
	if(ip < 10000) lcnt[ip]++;
}

void report(void){
	int i, s;
	s = 0;
	for(i=0; i<EOC; i++)
		if(icnt[i] != 0){
			prCode2(stdout,i);
			printf(" %d\n",icnt[i]);
			s += icnt[i];
		}
	printf("total %d\n",s);

	s = 0;
	for(i=0; i<10000; i++)
		if(lcnt[i] != 0){
			s += lcnt[i];
			printf("%d %d\n",i,lcnt[i]);
		}
	printf("total %d\n",s);

}

#endif
*/
void initfile(void){
	int i, a;
	filep[0] = stdin;
	filep[1] = stdout;
	filep[2] = stderr;
	for(i=3; i<MAXFILE; i++)
		filep[i] = NULL;
	// init small constants
	a = 390;
	for(i=-10; i<=300; i++,a++)
		M[a] = i;
}

// get a new file pointer, return 0 if full
PRIVATE int newfp(void){
	int i;
	for(i=3; i<MAXFILE; i++)
		if(filep[i]==NULL) return i;
	return 0;
}
/*
void dumpDS(void){
	int i;
	for(i=1; i<=10; i++)
		printf("%d:%d ",i,M[i]);
	nl();
	for(i=64; i<70; i++)
		printf("%d:%d ",i,M[i]);
	printf("\n");
}

PRIVATE void trace(void){
	int i, a, b;
	printf("%5d ",ip);
	prCode(stdout,ip);
	nl();
	dumpDS();
	printf("fp:%d ",fp);
	a = fp - 10;
	if( a < 0 ) a = 0;
	for(i=a; i<=fp+5; i++)
		if(i == fp)
			printf("[%d] ",SS[i]);
		else
			printf("%d ",SS[i]);
	printf("\n");
}

PRIVATE void error(char *mess, int i){
	printf("%s at ip %d\n",mess,i);
	printf("noi %d\n",noi);
	exit(0);
}
*/
// static alloc memory of size a
int xalloc(int a){
	int b;
	b = DS;
	if( DS + a >= MAXDS )
		error("alloc: out of memory");
	DS += a;
	return b;
}

// gets() and put it in M[ads]
PRIVATE void xgets(FILE *fp, int ads){
	char buf[MAXLBUF], *c;
	int i;
	c = fgets(buf,MAXLBUF,fp);
	if( c != NULL ){
		for(i=0; buf[i] != 0; i++, ads++)
			M[ads] = (int)buf[i] & 255;
		M[ads] = 0;
	}else{
		M[ads] = EOF_CHAR;
		M[ads+1] = 0;
	}
}

// save fp, and 1..n register
PRIVATE void saveContext(int n){
	int i, k;
	k = sp + 1;
	for(i=1; i<=n; i++){
		SS[k] = M[i];
		k++;
	}
	SS[k] = fp;
	fp = k;
	sp = k;
}

PRIVATE void restoreContext(int n){
	int i, k;
	k = fp - 1;
	for(i=n; i>=1; i--){
		M[i] = SS[k];
		k--;
	}
	sp = k;
	fp = SS[fp];
}

// exec syscall n, options in: a, b
PRIVATE void xsys(int n, int a, int b){
	int f;
	char name[NAMELEN];

	switch(n){
	case 1: printf("%d",a); break;		// pr int
	case 2: printf("%c",a); break;		// pr char
	case 3: M[RETVAL]=getchar(); break;	// getchar
	case 4: xgets(stdin,a); break;		// gets
	case 5:								// fopen fname mode
		strunpack(name,a);				// file name
		f = newfp();
		if(f == 0) error("cannot open, too many files");
		switch(b){
		case 0: filep[f] = fopen(name,"r"); break;
		case 1: filep[f] = fopen(name,"w"); break;
		case 2: filep[f] = fopen(name,"rw"); break;
		default: error("unknown file mode");
		}
		M[RETVAL] = f;					// return file pointer
		break;
	case 6:								// fclose fp
		fclose(filep[a]);
		filep[a] = NULL;
		break;
	case 7:	fprintf(filep[a],"%d",b); break; 	// fpr fp num
	case 8:	fprintf(filep[a],"%c",b); break; 	// fpr fp c
	case 9:	M[RETVAL] = fgetc(filep[a]); break; // fgetc fp
	case 10: xgets(filep[a],b); break;			// fgets fp buf
	case 11:									// fprints fp str
		strunpack(strbuf,b);
		fprintf(filep[a],"%s",strbuf);
		break;
	case 12:					// eval from som
		user_cs();				// switch to user space
		saveContext(3);			// fix 3 reg, should use b
		eval(a);
		restoreContext(3);
		sys_cs();				// switch back
		break;

//	case 13:  runflag = 0;  	// halt, this is in eval()

	case 14: M[RETVAL] = xalloc(a);	break;
	case 15:					// exec load from user space
		sys_cs();
		M[1] = a;				// filename
		eval(call_load);
		user_cs();
		break;
	case 16: M[RETVAL] = lex(filep[a]); break; // lex
	default: error("undefined syscall");
	}
}

// instruction format tx-code   op c a b

#define dec(a,b)	a=CS[ip+1]; b=CS[ip+2]
#define deca(a)     a=CS[ip+1]
#define nxt()		ip+=3
#define bop(op)		dec(a,b); M[c]=M[a] op M[b]; nxt()
#define jc(op)		dec(a,b); if(M[a] op M[b]) ip+=c; else nxt()

void eval(int ip){
	int runflag;
	int i, arity, nlocal, op;
	int a2, c, a, b;

	runflag = 1;
	while(runflag ) {
//		noi++;
		op = CS[ip] & 255;
		c = CS[ip] >> 8;
/*
		#ifdef PROF
			stat(ip);
		#endif
*/
// ---- start exec -----

	switch(op){
	case tAdd: bop(+); break;
	case tSub: bop(-); break;
	case tMul: bop(*); break;
	case tDiv: bop(/); break;
	case tMod: bop(%); break;
	case tAnd: bop(&); break;
	case tOr:  bop(|); break;
	case tXor: bop(^); break;
	case tShl: bop(<<); break;
	case tShr: bop(>>); break;
	case tEq:  bop(==); break;
	case tNe:  bop(!=); break;
	case tLt:  bop(<);  break;
	case tLe:  bop(<=); break;
	case tGt:  bop(>);  break;
	case tGe:  bop(>=); break;

	case tLdx:
		dec(a,b);
		M[c] = M[ M[a] + M[b] ];
		nxt();
		break;
	case tStx:
		dec(a,b);
		M[ M[a] + M[b] ] = M[c];
		nxt();
		break;
	case tJmp: ip += c; break;
	case tJf:
		deca(a);
		if(M[a] == FALSE) ip += c; else nxt();
		break;
	case tJt:
		deca(a);
		if(M[a] != FALSE) ip += c; else nxt();
		break;
	case tJeq: 	jc(==); break;
	case tJne: 	jc(!=); break;
	case tJlt: 	jc(<);  break;
	case tJle: 	jc(<=); break;
	case tJgt:	jc(>);  break;
	case tJge:	jc(>=); break;

	case tCall:					// fun.arity fs
		dec(a,b);				// b is ads, a 2nd param
		arity = CS[b] >> 8;
		nlocal = CS[b+1];
		if (fp + nlocal + FUDGE > MAXSS)
			error("stack overflow");
		// move all parameters to param[.]
		switch(arity){
		case 0: break;
		case 1: param[1] = M[c]; break;
		case 2:
			param[1] = M[c];
			param[2] = M[a];
			break;
		default:				// more than 2
			param[1] = M[c];
			param[2] = M[a];
			for(i=3; i<=arity; i++)
				param[i] = SS[sp-arity+i];
			sp = sp-arity+2;
		}
		// save reg, new frame is at sp
		for(i=1; i<=nlocal; i++)
			SS[sp+i] = M[i];
		for(i=1; i<=arity; i++)	// transfer param
			M[i] = param[i];
		c = sp + nlocal + 2;	// new fp
		SS[c-1] = ip + 3;		// ret ads
		SS[c] = fp;				// fp'
		fp = c;
		ip = b + 3;
		sp = fp;
		break;
	case tRet:					// ret.v fs
		deca(a);
		sp = fp - a - 2;
		M[RETVAL] = M[c];		// ret value
		for(i=1; i<=a; i++)		// restore reg
			M[i] = SS[sp+i];
		ip = SS[fp-1];			// restore ip
		fp = SS[fp];			// restore fp
		break;

	case tCallt:
		dec(a,b);
		arity = CS[b] >> 8;
		switch(arity){
		case 0: break;
		case 1: param[1] = M[c]; break;
		case 2:
			param[1] = M[c];
			param[2] = M[a];
			break;
		default:				// more than 2
			param[1] = M[c];
			param[2] = M[a];
			for(i=3; i<=arity; i++)
				param[i] = SS[sp-arity+i];
			sp = sp-arity+2;
		}
		for(i=1; i<=arity; i++)	// transfer param
			M[i] = param[i];
		ip = b + 3;
		break;
	case tNot:
		deca(a);
		M[c] = M[a] == FALSE;
		nxt();
		break;
	case tMov:
		deca(a);
		M[c] = M[a];
		nxt();
		break;
	case tPush:
		sp++;
		SS[sp] = M[c];
		nxt();
		break;
	case tEfor:
		dec(a,b);
		if(++M[a] <= M[b]) ip += c;
		else nxt();
		break;
	case tCase:
		dec(a,b);					// a lo, b hi
		a2 = M[c];
		if( a2 >= a && a2 <= b )
			ip += CS[ip+3+a2-a];	// jump vector
		else
			ip += (b-a+4);			// end of jtab
		break;
	case tSys:
		dec(a,b);
		if(c == 13) runflag = 0;
		else xsys(c,M[a],M[b]);
		nxt();
		break;

	default:
		error("undefined opcode");
	}

// ------ end exec --------

	}
//	printf("noi %d\n",noi);
}



